﻿using System;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Threading;

namespace P2PQRelay
{
    public class ReceivedEventArgs : EventArgs
    {
        public string Message { get; set; }
    }

    public class ConnectionFailedEventArgs : EventArgs
    {
        public Exception Exception { get; set; }
    }
    
    public class Client
    {
        public event EventHandler<ReceivedEventArgs> OnReceived;
        public event EventHandler OnDisconnected;
        public event EventHandler OnConnected;
        public event EventHandler<ConnectionFailedEventArgs> OnConnectionFailed;

        private volatile bool isRunning = false;
        private AutoResetEvent stoppedSignal;
        private Thread thread;

        private bool connected = false;

        public IPAddress Address { get; set; }
        public int Port { get; set; }
        public bool IsConnected { get { return this.connected; } }

        public Client()
        {
            this.stoppedSignal = new AutoResetEvent(false);
        }

        public void Start()
        {
            if (this.isRunning)
            {
                return;
            }

            this.connected = false;
            this.thread = new Thread(new ThreadStart(Run));
            this.thread.Start();
        }

        public void Stop()
        {
            this.isRunning = false;
            if(null != this.thread)
            {
                if(this.thread.IsAlive)
                {
                    this.stoppedSignal.WaitOne();
                    this.thread = null;
                }
            }
            this.connected = false;
        }

        void Run()
        {
            TcpClient client = new TcpClient(AddressFamily.InterNetwork);
            try
            {
                client.Connect(this.Address, this.Port);
            }
            catch (Exception e)
            {
                if (null != this.OnConnectionFailed)
                {
                    ConnectionFailedEventArgs args = new ConnectionFailedEventArgs();
                    args.Exception = e;
                    OnConnectionFailed.BeginInvoke(this, args, (AsyncCallback)(ar => {
                        ((EventHandler<ConnectionFailedEventArgs>)ar.AsyncState).EndInvoke(ar);
                    }), OnConnectionFailed);
                }
                return;
            }

            NetworkStream ns = client.GetStream();

            try
            {
                beginRead(ns);
            }
            catch(Exception e)
            {
                System.Diagnostics.Debug.WriteLine(e);
                if (null != this.OnDisconnected)
                {
                    OnDisconnected.BeginInvoke(this, null, (AsyncCallback)(ar =>
                    {
                        ((EventHandler)ar.AsyncState).EndInvoke(ar);
                    }), OnDisconnected);
                }
            }
            finally
            {
                ns.Close();
            }

            this.isRunning = false;
            this.stoppedSignal.Set();
        }

        void beginRead(NetworkStream ns)
        {
            byte[] readBuf = new byte[1024];
            byte[] sendBuf = Encoding.UTF8.GetBytes("P2PQRelay/1.0 Heartbeat");
            const int READCOMPLETE_TIMEOUT = 100;
            const int HEARTBEAT_INTERVAL = 5000 / READCOMPLETE_TIMEOUT;
            int heartbeat_interval_count = 0;
            const int HEARTBEAT_SEND_TIMEOUT = 1000;
            int heartbeat_norecv_count = 0;
            const int HEARTBEAT_NORECV_TIMEOUT = 10000 / READCOMPLETE_TIMEOUT;

            this.isRunning = true;

            while (this.isRunning)
            {
                using (MemoryStream ms = new MemoryStream())
                {
                    do
                    {
                        ManualResetEvent readCompleteSignal = new ManualResetEvent(false);
                        IAsyncResult reading = ns.BeginRead(readBuf, 0, readBuf.Length, (AsyncCallback)(ar =>
                        {
                            ManualResetEvent signal = (ManualResetEvent)ar.AsyncState;
                            if (ar.IsCompleted && ar.CompletedSynchronously)
                            {
                                signal.Set();
                            }
                        }), readCompleteSignal);

                        if (!this.connected)
                        {
                            this.connected = true;
                            OnConnected.BeginInvoke(this, null, (AsyncCallback)(ar =>
                            {
                                ((EventHandler)ar.AsyncState).EndInvoke(ar);
                            }), OnConnected);
                        }

                        do
                        {
                            if (!this.isRunning)
                            {
                                if (null != this.OnDisconnected)
                                {
                                    OnDisconnected.BeginInvoke(this, null, (AsyncCallback)(ar => {
                                        ((EventHandler)ar.AsyncState).EndInvoke(ar); 
                                    }), OnDisconnected);
                                }

                                return;
                            }
                            readCompleteSignal.WaitOne(READCOMPLETE_TIMEOUT);

                            if (++heartbeat_norecv_count >= HEARTBEAT_NORECV_TIMEOUT)
                            {
                                // 一定時間ハートビート返送を受信しなかった
                                if (null != this.OnDisconnected)
                                {
                                    OnDisconnected.BeginInvoke(this, null, (AsyncCallback)(ar =>
                                    {
                                        ((EventHandler)ar.AsyncState).EndInvoke(ar);
                                    }), OnDisconnected);
                                }
                                System.Diagnostics.Debug.WriteLine("P2PQRelayClient.beginRead: heartbeat dead");
                                return;
                            }

                            if(++heartbeat_interval_count >= HEARTBEAT_INTERVAL)
                            {
                                // ハートビート送信
                                ManualResetEvent sendCompleteSignal = new ManualResetEvent(false);
                                ns.BeginWrite(sendBuf, 0, sendBuf.Length, (AsyncCallback)(ar =>
                                {
                                    ManualResetEvent signal = (ManualResetEvent)ar.AsyncState;
                                    if (ar.IsCompleted)
                                    {
                                        signal.Set();
                                    }
                                }), sendCompleteSignal);
                                if(!sendCompleteSignal.WaitOne(HEARTBEAT_SEND_TIMEOUT))
                                {
                                    // 送信タイムアウト
                                    if (null != this.OnDisconnected)
                                    {
                                        OnDisconnected.BeginInvoke(this, null, (AsyncCallback)(ar =>
                                        {
                                            ((EventHandler)ar.AsyncState).EndInvoke(ar);
                                        }), OnDisconnected);
                                    }
                                    System.Diagnostics.Debug.WriteLine("P2PQRelayClient.beginRead: timed out on sending heartbeat");
                                    return;
                                }
                                heartbeat_interval_count = 0;
                            }

                        } while (!reading.IsCompleted);

                        int size = ns.EndRead(reading);

                        if (0 == size)
                        {
                            // 切断された
                            if (null != this.OnDisconnected)
                            {
                                OnDisconnected.BeginInvoke(this, null, (AsyncCallback)(ar => {
                                    ((EventHandler)ar.AsyncState).EndInvoke(ar); 
                                }), OnDisconnected);
                            }
                            return;
                        }
                        ms.Write(readBuf, 0, size);

                    } while (ns.DataAvailable);

                    string msg = Encoding.UTF8.GetString(ms.ToArray());
                    //System.Diagnostics.Debug.WriteLine(msg);

                    const string RELAY_TEXT = "P2PQRelay/1.0 Relay ";
                    const string HEARTBEAT_TEXT = "P2PQRelay/1.0 Heartbeat OK";
                    if (msg.StartsWith(RELAY_TEXT))
                    {
                        if (null != this.OnReceived)
                        {
                            ReceivedEventArgs args = new ReceivedEventArgs();
                            args.Message = msg.Substring(RELAY_TEXT.Length);

                            OnReceived.BeginInvoke(this, args, (AsyncCallback)(ar =>
                            {
                                ((EventHandler<ReceivedEventArgs>)ar.AsyncState).EndInvoke(ar);
                            }), OnReceived);
                        }
                    }
                    else if (msg == HEARTBEAT_TEXT)
                    {
                        //System.Diagnostics.Debug.WriteLine("heartbeat reply received");
                        heartbeat_norecv_count = 0;
                    }
                }
            }
        }
    }
}
